/////////////////////////////////////////////////////////////////////
// Mismatch provides methods for analyzing coalescent trees to
// determine the number of mutations distinguishing each pair of
// terminal nodes. These can be used to determine the mismatch
// distribution (also called the distribution of pairwise
// differences). They can also be used to calculate the average number
// of differences across all pairs, or mean pairwise difference (MPD).

public class Mismatch
{

    int sampleSize;
    int[] mismatch;
    Node[] tree;
    int maxPairwiseDiff = 10000;

    public Mismatch(Tree t) {
        this.tree = t.tree;

        sampleSize = (tree.length - 1) / 2 + 1;

	generateMismatch();

    }

    // Follow the tree from node to node, tabulating the number
    // of differences between each pair
    private void generateMismatch() {

	Node node = null;

	// Initialize mismatch, which is a class variable
	// that will hold the mismatch distribution
	mismatch = new int[maxPairwiseDiff];

	// Recursively crawl the tree, counting the number of
	// differences between each pair of terminal nodes
	for (int j = 0; j < sampleSize; j++){
	    visit(tree[j], 0);
	    for (int k = 0; k < 2 * sampleSize - 1; k++)
		tree[k].visited=0;
	}

	// Correct the mismatch distribution by eliminating self
	// comparisons (which always have 0 differences) and repeat
	// comparisons (order doesn't matter, so comparing A and B is
	// the same as comaparing B with A.
	mismatch[0] -= sampleSize;
  	for (int j = 0; j < maxPairwiseDiff; j++){
  	    mismatch[j]/=2;
  	}
    }

    // Given a node, mark it as visited, add the number of mutations
    // distinguishing the node and its ancestor, and call the godown
    // method to examine the node's ancestor's ancestors. This
    // process is recursive.
    private int visit(Node testnode, int total) {

	// Mark the node as visited
	testnode.visited = 1;

	// Visit the node's descendants
	godown(testnode, total);
	
	// If the node doesn't have any ancestors the
	// total is returned
	if (testnode.ancestor == null)
	    return(total);

	// Visit the node's ancestor and add its mutations to
	// the total	
	total += visit(testnode.ancestor, total+testnode.nMuts);

	return(total);
    }

    // Given a node, check whether it has any descendants. If no, add
    // 1 to the k differences row in the mismatch distribution. If
    // yes, examine the node's descendants. This process is recursive.
    private void godown(Node testnode, int total) {
	if (testnode.desc1 == null) {
	    if (total > maxPairwiseDiff-1)
		mismatch[maxPairwiseDiff-1]++;
	    else mismatch[total]++;
	    return;
	}

	if (((testnode.desc1).visited)==0) {
	    godown(testnode.desc1, total + (testnode.desc1).nMuts);
	}

	if (((testnode.desc2).visited)==0) {
	    godown(testnode.desc2, total + (testnode.desc2).nMuts);
	}
	
	return;
    }

    // Return the MPD after calculating it from the mismatch
    // distribution
    public double getMPD() {
	double mpd = 0;
	double comparisons = (sampleSize * sampleSize - sampleSize) / 2;
	int tot = 0;
	int lastEntry = maxPairwiseDiff - 1;

	while (mismatch[lastEntry] == 0)
	    lastEntry--;
	
	for (int i = 0; i <= lastEntry; i++) {
	    mpd += (double)mismatch[i] / (double)comparisons * i;
	    tot += mismatch[i];
	}
	return mpd;
    }

}
