import java.text.*;

public class STUSC
{
    public static void main( String[] args ) {

        int n;               // Sample size
        double mpd;          // Mean pairwise difference
        int segSites;        // Segregating sites
        int sgltons;         // Singletons
	String historyFile;  // Input history filename
        int its;             // Iterations
        boolean genDistribs; // Output simulated distributions, or not
	int outputVerbosity; // Specify how much output is needed

	// CLParser is a simple command line parser
        CLParser parser = new CLParser(args);
       	STUSC ntusc = new STUSC();

	// Retrieve parameters from parser
        n = parser.getN();
        segSites = parser.getSegSites();
        mpd = parser.getMPD();
        sgltons = parser.getSgltons();
	historyFile = parser.getHistoryFile();
        its = parser.getIts();
        genDistribs = parser.getGenDistribs();
	outputVerbosity = parser.getOutputVerbosity();

	// Get the population history

	History epochs = new History();
	epochs.readHistory(historyFile);

	// Pass parameters to doAnalysis, which carries out a
	// the statistical tests, summarizes the output, and
	// returns it as a string
	String result = ntusc.doAnalysis(
					n,
					segSites,
					mpd,
					sgltons,
					epochs,
					its,
					genDistribs,
					outputVerbosity);
	System.out.println(result);

    }

    ///////////////////////////////////////////////////
    //
    // doAnalysis is the core of STUSC. It takes input data
    // and processing parameters, and uses them to calculate
    // statistics capable of detecting natural selection,
    // simulate coalescent trees and  statistics, and
    // tabulates p-values. It returns a String containing
    // the results of the tests in one of two selected
    // formats.

    public String doAnalysis (int n,
			      int segSites,
			      double mpd,
			      int sgltons,
			      History epochs,
			      int its,
			      boolean genDistribs,
			      int outputVerbosity) {

	// Tests contains methods for calculating statistics using
	// input data, such as number of segregating sites, mean
	// pairwise difference, and number of singletons.
	Tests stat = new Tests();

	// Calculate values for the user's data
	double obsTajD  = stat.tajD(n, mpd, segSites);
	double obsFuLiD  = stat.fuLiD(n, sgltons, segSites);
	double obsFuLiDs = stat.fuLiDstar(n, sgltons, segSites);
	double obsFuLiF   = stat.fuLiF(n, mpd, sgltons, segSites);
	double obsFuLiFs  = stat.fuLiFstar(n, mpd, sgltons, segSites);

	// Initialize simulated values to 0.0
	int simS = 0;
	int sims = 0;
	double simMpd = 0.0;
	double simTajD  = 0.0;
	double simFuLiD  = 0.0;
	double simFuLiDs = 0.0;
	double simFuLiF   = 0.0;
	double simFuLiFs  = 0.0;

	// Initialize p-values to 0.0
	double TajDPValue = 0.0;
	double FuLiDPValue = 0.0;
	double FuLiDsPValue = 0.0;
	double FuLiFPValue  = 0.0;
	double FuLiFsPValue  = 0.0;


	////////////////////////////////////////////////////////
	// The different statistics require different input.
	// For example, Tajima's D can't be calulated without
	// the mean pairwise difference and Fu and Li's D can't
	// be calculated without singletons

	boolean doTajD = true;
	boolean doFuLiD = true;
	boolean doFuLiF = true;

	// Tajima's D and Fu and Li's F/F* require mpd
	if (mpd == -1.0) {
	    doTajD = false;
	    doFuLiF = false;
	}

	// Fu and Li's D/D*/F/F* require singletons
	if (sgltons == -1) {
	    doFuLiD = false;
	    doFuLiF = false;
	}

	//
	////////////////////////////////////////////////////////


	// Initialize output, which will hold a string indicating
	// the results of simulations and tests
	String output = new String("");

	// A header labeling the columns of simulated statistics
	if (genDistribs) {
	    output += "# Tajima's D, Fu and Li's D,"
		+ "Fu and Li's D*, "
		+ "Fu and Li's F,  Fu and Li's F*\n";
	}

	////////////////////////////////////////////////////////
	// Using the input parameters and initialized variables,
	// simulate coalescent trees and use them to calculate
	// simulated values of Tajima's D, Fu and Li's D, etc.
	// If the user has asked for the  simulated values,
	// output them on the fly. Also, cumulatively calculate
	// p-values.
	for (int i = 0; i < its; i++) {
	
	    // This is a coalescent tree
	    Tree ct = new Tree();
	    ct.setParams(n, epochs);
	    ct.genTopology();
	    ct.genIntervals();
	    ct.genMutations();

	    // Determine the site frequency spectrum derived
	    // from the simulated tree
	    Spectrum spec = new Spectrum(ct);

	    // Check the spectrum to make sure there is at least one
	    // segregating site. If there are no segregating sites
	    // the neutrality tests are nonsensical so restart the
	    // loop.
	    if (spec.getSegSites() == 0) {
		i--;
		continue;
	    }

	    // Mismatch distribution derived from the simulated
	    // tree
	    Mismatch mism = new Mismatch(ct);    

	    // Calculate simulated values derived from the tree
	    simTajD = stat.tajD(spec, mism);
	    simFuLiD = stat.fuLiD(spec);
	    simFuLiDs = stat.fuLiDstar(spec);
	    simFuLiF = stat.fuLiF(spec, mism);
	    simFuLiFs = stat.fuLiFstar(spec, mism);
	    
	    // If the user wants the entire list of simulated
	    // values, which can be used to calculate distributions,
	    // append them to output
	    if (genDistribs) {

		simS = spec.getSegSites();
		sims = spec.getSingletons();
		simMpd = mism.getMPD();

		output += 
		    // simS                     + ","
		    // + sims                   + ","
		    // formatNumber(simMpd)   + ","
		      formatNumber(simTajD)   + ","
		    + formatNumber(simFuLiD)  + ","
		    + formatNumber(simFuLiDs) + ","
		    + formatNumber(simFuLiF)  + ","
		    + formatNumber(simFuLiFs);

		if (i < its - 1) {
		    output += "\n";
		}
	    }


	    // Incrementally increase p-values
	    if (simTajD > obsTajD) {
		TajDPValue += 1.0/(double)its;
	    }

	    if (simFuLiD > obsFuLiD) {
		FuLiDPValue += 1.0/(double)its;
	    }

	    if (simFuLiDs > obsFuLiDs) {
		FuLiDsPValue += 1.0/(double)its;
	    }

	    if (simFuLiF > obsFuLiF) {
		FuLiFPValue += 1.0/(double)its;
	    }

	    if (simFuLiFs > obsFuLiFs) {
		FuLiFsPValue += 1.0/(double)its;
	    }

	}
	//
	////////////////////////////////////////////////////////

	// If the user wants tidy, verbose output...
	if (outputVerbosity == 0) {
	    output += verboseResults (n,
				 segSites,
				 mpd,
				 sgltons,
				 epochs,
				 its,
				 obsTajD,
				 TajDPValue,
				 obsFuLiD,
				 FuLiDPValue,
				 obsFuLiDs,
				 FuLiDsPValue,
				 obsFuLiF,
				 FuLiFPValue,
				 obsFuLiFs,
				 FuLiFsPValue,
				 doTajD,
				 doFuLiF,
				 doFuLiD);
	} else if (outputVerbosity == 1) {
	    // If the user wants csv output...
	    output += csvResults (obsTajD,
				   TajDPValue,
				   obsFuLiD,
				   FuLiDPValue,
				   obsFuLiDs,
				   FuLiDsPValue,
				   obsFuLiF,
				   FuLiFPValue,
				   obsFuLiFs,
				   FuLiFsPValue,
				   doTajD,
				   doFuLiF,
				   doFuLiD);
	} else {
	    // No output from the analysis
	}

	return output;

    }

    //
    ///////////////////////////////////////////////////

    ///////////////////////////////////////////////////
    //
    // checkParams verifies that the user input is adequate
    // and correct
    //

    public void checkParams (
			     int n,
			     int segSites,
			     double theta0,
			     double tau,
			     double growth,
			     int its
			     ) {

	int parameterError = 0;

	if (n == -1) {
	    System.out.println("Specify sample size using -n.");
	    parameterError = 1;
	}

	if (segSites == -1) {
	    System.out.println("Specify segregating sites size using -S.");
	    parameterError = 1;
	}

	if (theta0 == -1.0) {
	    System.out.println("Specify theta0 using -q.");
	    parameterError = 1;
	}

	if (tau == -1.0) {
	    System.out.println("Specify tau using -t.");
	    parameterError = 1;
	}

	if (growth == -1) {
	    System.out.println("Specify growth factor using -f.");
	    parameterError = 1;
	}

	if (its == -1) {
	    System.out.println("Specify iterations using -i.");
	    parameterError = 1;
	}

	if (parameterError == 1) {
	       System.exit(1);
	}

    }

    //
    ///////////////////////////////////////////////////

    ///////////////////////////////////////////////////
    //
    // formatNumber formats a double into a standard format
    //

    public String formatNumber(double d) {

	DecimalFormat formatter = new DecimalFormat("###.####");
	formatter.setMinimumIntegerDigits(1);
	formatter.setMinimumFractionDigits(4);
	String formattedNumber = formatter.format(d);

	return formattedNumber;
    }

    //
    ///////////////////////////////////////////////////
	
    ///////////////////////////////////////////////////
    //
    // verboseResults outputs results on two lines. The
    // first is a comment line labeling the columns.
    // The second is the observed values and p-values
    // for each statistic
    //

    public String csvResults (double obsTajD,
				double TajDPValue,
				double obsFuLiD,
				double FuLiDPValue,
				double obsFuLiDs,
				double FuLiDsPValue,
				double obsFuLiF,
				double FuLiFPValue,
				double obsFuLiFs,
				double FuLiFsPValue,
				boolean doTajD,
				boolean doFuLiF,
				boolean doFuLiD) {

	String output = "";

	output += "#";

	if (doTajD) {

	    output += "Tajima's D,p(Greater Tajima's D)";

	    if (doFuLiD || doFuLiF)
		output += ",";
	}

	if (doFuLiD) {

	    output += "Fu and Li's D,p(Greater Fu and Li's D),Fu "
		+ "and Li's D*,p(Greater Fu and Li's D*)";

	    if (doFuLiF)
		output += ",";
	}

	if (doFuLiF)

	    output += "Fu and Li's F,p(Greater Fu and Li's F),Fu "
		+ "and Li's F*,p(Greater Fu and Li's F*)";

	output += "\n";

	if (doTajD) {
	    output += formatNumber(obsTajD) + ","
		+ formatNumber(TajDPValue);

	    if (doFuLiD || doFuLiF)
		output += ",";
	}

	if (doFuLiD) {
	    output += formatNumber(obsFuLiD) + ","
		+ formatNumber(FuLiDPValue) + ","
		+ formatNumber(obsFuLiDs) + ","
		+ formatNumber(FuLiDsPValue);
	    if (doFuLiF)
		output += ",";
	}

	if (doFuLiF)
	    output += formatNumber(obsFuLiF) + ","
		+ formatNumber(FuLiFPValue) + ","
		+ formatNumber(obsFuLiFs) + ","
		+ formatNumber(FuLiFsPValue);

	return output;

    }

    //		
    ///////////////////////////////////////////////////

    ///////////////////////////////////////////////////
    //
    // verboseResults outputs an easily readable summary
    // of the input data and output results
    //

    public String verboseResults (int n,
				  int segSites,
				  double mpd,
				  int sgltons,
				  History epochs,
				  int its,
				  double obsTajD,
				  double TajDPValue,
				  double obsFuLiD,
				  double FuLiDPValue,
				  double obsFuLiDs,
				  double FuLiDsPValue,
				  double obsFuLiF,
				  double FuLiFPValue,
				  double obsFuLiFs,
				  double FuLiFsPValue,
				  boolean doTajD,
				  boolean doFuLiF,
				  boolean doFuLiD) {

	String output = "";

	///////////////////////////////////////////////////
	//
	// Output: Input data, parameters, and notes
	//

	output += "\nSTUSC version 1.0" + "\n";
	output += "=================" + "\n\n";

	output += "  Data:" + "\n";
	output += "\t     n = " + n + " haploid genomes" + "\n";
	if (doFuLiD)
	    output += "\t     s = " + sgltons + " singletons" + "\n";
	output += "\t     S = " + segSites + " segregating sites" + "\n";
	if (doTajD)
	    output += "\t   mpd = " + mpd + " nucleotides\n";
	output += "    Iterations = " + its + "\n";

	output += "\n";


	output +=     "  History:\tEpoch\tTheta\tTau\n";
	for (int i = 0; i < epochs.getNumEpochs()-1; i++)
	    output += "           \t" + (i+1) + "\t"
		+ epochs.getTheta(i) + "\t" + epochs.getTau(i) + "\n";

	output += "           \t" + epochs.getNumEpochs() + "\t"
	    + epochs.getTheta(epochs.getNumEpochs()-1) + "\tInf.\n";

	output += "\n";

	//
	///////////////////////////////////////////////////


	///////////////////////////////////////////////////
	//
	// Output: Observed statistics and p-values
	//

	output += "  Statistics and Tests:\n";

	if (doTajD) {
	    output += "     \tTajima's D     = ";

	    if (obsTajD > 0)
		output += " "; // Extra space to align w/ negative values
	    output += formatNumber(obsTajD) + "\tp(greater) = "
		+ formatNumber(TajDPValue) + "\n";
	}

	if (doFuLiD) {
	    output += "     \tFu and Li's D  = ";

	    if (obsFuLiD > 0)
		output += " ";  // Extra space to align w/ negative values
	    output += formatNumber(obsFuLiD) + "\tp(greater) = "
		+ formatNumber(FuLiDPValue) + "\n     \tFu and Li's D* = ";

	    if (obsFuLiDs > 0)
		output += " ";// Extra space to align w/ negative values
	    output += formatNumber(obsFuLiDs) + "\tp(greater) = "
		+ formatNumber(FuLiDsPValue) + "\n";
	}

	if (doFuLiF) {
	    output += "     \tFu and Li's F  = ";
	    if (obsFuLiF > 0)
		output += " ";// Extra space to align w/ negative values
	    output += formatNumber(obsFuLiF) + "\tp(greater) = "
		+ formatNumber(FuLiFPValue) + "\n";

	    output += "     \tFu and Li's F* = ";
	    if (obsFuLiFs > 0)
		output += " ";// Extra space to align w/ negative values
	    output += formatNumber(obsFuLiFs)  + "\tp(greater) = "
		+ formatNumber(FuLiFsPValue) + "\n";
	}

	//
	///////////////////////////////////////////////////

	///////////////////////////////////////////////////
	//
	// Output: Notes on interpreting p-values
	//

	output += "\n\t"
	    + "p(greater) = Fraction of simulated values exceeding the\n\t"
	    + "             observed value.\n";
	output += "\n"
	    + "  Interpretation: A p(greater) below your chosen significance\n"
	    + "\tthreshold (e.g. 0.05) indicates that your observed value is\n"
	    + "\tsignificantly GREATER THAN expected under the assumed\n"
	    + "\tpopulation history. A p(greater) above your chosen\n"
	    + "\tsignificance threshold (e.g., 0.95) indicates that your\n"
	    + "\tobserved value is significantly LESSER THAN expected under\n"
	    + "\tthe assumed population history.\n";

	return output;

	//
	///////////////////////////////////////////////////

    }

}
