////////////////////////////////////////////////////////////////////
// Spectrum provides methods for analyzing coalescent trees to
// determine the number of mutants observed 1 times, 2 times, ..., k
// times in the Tree.

public class Spectrum
{

    int sampleSize;
    Node[] tree;
    int[] spectrum;
    

    public Spectrum(Tree t) {
	this.tree = t.tree;

	sampleSize =(tree.length - 1) / 2 + 1;

	this.generateSpectrum();
    }

    public int getSampleSize() {
	return sampleSize;
    }

    // generateSpectrum recursively visits nodes and their
    // descendants, cumulatively determining the frequency of each
    // mutation (which is just the number of descendants of the node.
    private void generateSpectrum() {
	Node node = null;

	spectrum = new int[sampleSize - 1];

	// root node is tree[2 * sampleSize - 2]
	node = tree[2 * sampleSize - 2];

	// Fill spectrum by examining the root node,
	// then its descendants
	fillSpectrum(node.desc1, spectrum);
	fillSpectrum(node.desc2, spectrum);

    }

    // Recursively visit the descendant nodes, updating the spectrum
    // as mutations are encountered
    private void fillSpectrum(Node node, int[] spectrum) {

	int nDesc = 0;
	
	if (node.desc1 == null) {
	    spectrum[0] += node.nMuts;
	} else {

	    nDesc = countDesc(node);
	    spectrum[nDesc - 1] += node.nMuts;

	    fillSpectrum(node.desc1, spectrum);
	    fillSpectrum(node.desc2, spectrum);

	}
    }

    // countDesc calculates the number of descendants a node
    // has, which is the number of times the mutations on
    // the node are observed in the tree
    private int countDesc(Node node) {
	int sum = 0;

	if (node.desc1 == null) {
	    return(1);
	} else {
	    sum += countDesc(node.desc1);
	    sum += countDesc(node.desc2);
	    return(sum);
	}
    }

    // Total the categories to get the number of segregating sites
    public int getSegSites() {
	int S = 0;

	for (int i = 0; i < spectrum.length; i++)
	    S += spectrum[i];

	return S;
    }

    // Get the number of singletons (the number in bin 0 of the spectrum)
    public int getSingletons() {
	return spectrum[0];
    }

    // Output the spectrum in a human readable format
    public void printSpectrum() {
	for (int i = 0; i < spectrum.length; i++)
	    System.out.println((i+1) + ": " + spectrum[i]);
    }

}
