////////////////////////////////////////////////////////////////////
// Tests provides methods for calculating five statistics useful for
// testing hypotheses about the effects of natural selection in a
// gene. The statistics are: Tajima's D, Fu and Li's F, Fu and Li's
// F*, Fu and Li's D, and Fu and Li's D*. Complete decriptions of the
// statistics are provided by Tajima 1989 and Fu and Li 1993.

public class Tests
{

    // Calculate Fu and Li's F
    public double fuLiF(int sampleSize,     // Sample size
			double mpd,         // Mean pairwise difference
			int numSingletons,  // Singletons
			int segSites) {     // Segregating sites

	double n = sampleSize;
	double pin = mpd;
	double nu = segSites;
	double nue = numSingletons;
	double an = 0.0;


	// The next several lines calculate intermediate values needed to
	// obtain the full test statistic
	for (int k = 1; k <= n-1; k++)
	    an += 1.0/(double)k;

	double an1 = 0.0;
	for (int k = 1; k <= n; k++)
	    an1 += 1.0/(double)k;

	double bn = 0.0;
	for (int k = 1; k <= n-1; k++)
	    bn += 1.0/(double)Math.pow(k, 2.0);

	double cn = 2.0 * (n*an - 2.0*(n-1.0)) / ((n-1.0) * (n-2.0));

	double vF = (cn + (2.0*(Math.pow(n, 2.0) + n + 3.0)/(9.0*n*(n-1.0))) - 2.0/(n - 1.0)) / (Math.pow(an, 2.0) + bn);
	
	double uF = (1.0 + (n+1.0)/(3.0*(n-1.0)) - 4.0*((n+1.0)/Math.pow(n-1, 2)) * (an1 - (2.0*n)/(n+1.0)))/an -vF;

	double F = (pin - nue)/Math.sqrt(uF*nu + vF*Math.pow(nu, 2.0));

	return F;

    }

    // Calculate Fu and Li's F from the site frequency spectrum and
    // mismatch distributions, which can be used to obtain the needed
    // parameters (sample size, mpd, singletons, and segregating
    // sites.
    public double fuLiF(Spectrum spec, Mismatch mism) {

	return fuLiF(spec.getSampleSize(),
		     mism.getMPD(),
		     spec.getSingletons(),
		     spec.getSegSites());
    }

    // Calculate Fu and Li's F* statistic
    public double fuLiFstar(int sampleSize,     // Sample size
			    double mpd,         // Mean pairwise difference
			    int numSingletons,  // Singletons
			    int segSites) {     // Segregating sites

	double n = sampleSize;
	double pin = mpd;
	double nus = numSingletons;
	double nu = segSites;

        // The next several lines calculate intermediate values needed
	// to obtain the full test statistic

	double an = 0.0;
	for (int k = 1; k <= n-1; k++)
	    an += 1.0/(double)k;

	double an1 = 0.0;
	for (int k = 1; k <= n; k++)
	    an1 += 1.0/(double)k;

	double bn = 0.0;
	for (int k = 1; k <= n-1; k++)
	    bn += 1.0/Math.pow(k, 2.0);

	double cn = 2.0 * ((n*an - 2.0*(n-1.0)) / ((n-1.0) * (n-2.0)));

	double dn = cn +
	            (n-2.0)/Math.pow(n-1.0, 2.0) + 2.0/(n-1.0) *
	            (3.0/2.0-(2*an1-3.0)/(n-2.0) - 1.0/n);

	double vFs = (dn + (2.0*(Math.pow(n, 2.0) + n + 3.0)/(9.0*n*(n-1.0))) - 2.0/(n - 1.0) * (4.0*bn-6.0+8.0/n)) / (Math.pow(an, 2.0) + bn);
	
	double uFs = (n/(n-1.0) + (n+1.0)/(3.0*(n-1.0))
		     - 2.0*2.0/(n*(n-1.0)) + 2.0*((n+1.0)/Math.pow(n-1.0, 2.0))
		     * (an1 - (2.0*n)/(n+1.0))) / an - vFs;

       	double Fs = (pin - (n-1.0)/n * nus)/Math.sqrt(uFs*nu + vFs*Math.pow(nu, 2.0));

	return Fs;

    }

    // Calculate Fu and Li's F* from the site frequency spectrum and
    // mismatch distributions, which can be used to obtain the needed
    // parameters (sample size, mpd, singletons, and segregating
    // sites.
    public double fuLiFstar(Spectrum spec, Mismatch mism) {

	return fuLiF(spec.getSampleSize(),
		     mism.getMPD(),
		     spec.getSingletons(),
		     spec.getSegSites());

    }

    // Calculate Fu and Li's D
    public double fuLiD(int sampleSize,      // Sample size
			int numSingletons,   // Singletons
			int segSites) {      // Segregating sites

	double n = sampleSize;
	double S = segSites;
	double nu = segSites;
	double nue = numSingletons;


	// The next several lines calculate intermediate values needed
        // to obtain the full test statistic
	double an = 0.0;
	for (int k = 1; k <= n-1; k++)
	    an += 1.0/(double)k;

	double bn = 0.0;
	for (int k = 1; k <= n-1; k++)
	    bn += 1.0/(double)Math.pow(k, 2);

	double cn = 2.0 * ((n*an - 2.0*(n-1.0)) / ((n-1.0) * (n-2.0)));

	double vD = 1.0 + Math.pow(an, 2.0)/(bn + Math.pow(an, 2.0)) * (cn - (n+1.0)/(n-1.0));
	double uD = an - 1.0 - vD;

	double D = (nu - an*nue)/Math.sqrt(uD*nu + vD*Math.pow(nu, 2));

	if (S == 0)
	    D = 0;

	return D;

    }

    // Calculate Fu and Li's D from the site frequency spectrum, which
    // can be used to obtain the needed parameters (sample size,
    // singletons, and segregating sites.
    public double fuLiD(Spectrum spec) {

	return fuLiD(spec.getSampleSize(),
		     spec.getSingletons(),
		     spec.getSegSites());

    }

    // Calculate Fu and Li's D*
    public double fuLiDstar(int sampleSize, int numSingletons, int segSites) {

	double n = sampleSize;
	double S = segSites;
	double nu = segSites;
	double nus = numSingletons;

        // The next several lines calculate intermediate values needed
        // to obtain the full test statistic

	double an = 0.0;
	for (int k = 1; k <= n-1; k++)
	    an += 1.0/(double)k;

	double an1 = 0.0;
	for (int k = 1; k <= n-1+1; k++)
	    an1 += 1.0/(double)k;

	double bn = 0.0;
	for (int k = 1; k <= n-1; k++)
	    bn += 1.0/Math.pow(k, 2.0);

	double cn = 2.0 * ((n*an - 2.0*(n-1.0)) / ((n-1.0) * (n-2.0)));

	double dn = cn +
	            (n-2.0)/Math.pow(n-1.0, 2.0) + 2.0/(n-1.0) *
	            (3.0/2.0-(2*an1-3.0)/(n-2.0) - 1.0/n);

	double vDs =  (Math.pow(n/(n-1.0), 2.0) * bn 
		       + Math.pow(an, 2.0)*dn 
		       - 2.0 * ((n * an * (an+1.0)) / Math.pow(n-1.0, 2.0)))
	              / (Math.pow(an, 2.0) + bn);

	double uDs = n/(n-1.0) * (an - n/(n-1.0)) - vDs;

	double Ds = (n/(n-1.0) * nu - an*nus)/Math.sqrt(uDs * nu + vDs * Math.pow(nu, 2));

	if (S == 0)
	    Ds = 0;

	return Ds;

    }

    // Calculate Fu and Li's Dstar from the site frequency spectrum,
    // which can be used to obtain the needed parameters (sample size,
    // singletons, and segregating sites.
    public double fuLiDstar(Spectrum spec) {

	return fuLiDstar(spec.getSampleSize(),
			 spec.getSingletons(),
			 spec.getSegSites());

    }

    // Calculate Tajima's D
    public double tajD(int sampleSize, double mpd, int segSites) {

	int i;
	double a2;
	double b1;
	double b2;
	double c1;
	double c2;
	double a1;
	double e1;
	double e2;
	double D;

	double n = sampleSize;
	double s = segSites;
	double k = mpd;

        // The next several lines calculate intermediate values needed
        // to obtain the full test statistic

	a1 = 0.0; 
	a2 = 0.0; 

	for (i = 1; i <= n-1; i++) {
	    a1 += (1.0 / (double)i); 
	    a2 += (1.0 / (double)(i * i)); 
	}

	b1 = (n + 1.0) / (3.0 * (n - 1.0)); 
	b2 = (2.0 * (n * n + n + 3.0)) / (9.0 * n * (n - 1.0)); 

	c1 = b1 - (1.0 / a1); 
	c2 = b2 - ((n + 2.0) / (a1 * n)) + (a2 / (a1 * a1)); 


	e1 = c1 / a1; 
	e2 = c2 / ((a1 * a1) + a2); 

	D = (k - (s / a1)) / (Math.sqrt((e1 * s) + (e2 * s) * (s - 1.0)));

	if (s == 0)
	    D = 0.0;

	return D;

    }

    // Calculate Tajima's D from the site frequency spectrum,
    // which can be used to obtain the needed parameters (sample size,
    // mpd, and segregating sites.
    public double tajD(Spectrum spec, Mismatch mism) {

	return tajD(spec.getSampleSize(),
		       mism.getMPD(),
		       spec.getSegSites());

    }
}
