//////////////////////////////////////////////////////////////////////
// Tree provides methods for generating and storing a coalescent tree,
// including the tree's topology, the length of time between a pair of
// nodes and their common ancestor, and the number of mutations
// distinguishing a node and its ancestor.

public class Tree
{
    int sampleSize;   // The haploid population size
    History history;  // A population history
    int[] spectrum;   // Can hold a site frequency spectrum
    int[] mismatch;   // Can hold a mismatch distribution
    Node[] tree;      // Holds the tree, which is a list of conneced nodes

    // Set the parameters used to generate the tree
    public void setParams(int sampleSize,
			  History history)
    {
	int i;

	this.sampleSize = sampleSize;
	this.history    = history;

	tree = new Node[2 * sampleSize];

    }

    // For sampleSize number of nodes, randomly generate nodes
    // representing common ancestors.
    public void genTopology() {
	int in;
	int pick;
	Node[] list;

	list = new Node[sampleSize];

	for (in = 0; in < ( 2 * sampleSize - 1 ); in++)
	    tree[in] = new Node(in);

	for (in = 0; in < sampleSize; in++)
	    list[in] = tree[in];

	for (in = sampleSize; in > 1; in --) {

	    pick = (int)(Math.random() * in);
	    list[pick].ancestor = tree[2 * sampleSize - in];
	    tree[2 * sampleSize - in].desc1 = list[pick];
	    list[pick] = list[in - 1];

	    pick = (int)(Math.random() * (in - 1));
	    list[pick].ancestor = tree[2 * sampleSize - in];
	    tree[2 * sampleSize - in].desc2 = list[pick];
	    list[pick] = tree[2 * sampleSize - in];

	}
    }

    // Generate the times between pairs of nodes and their shared
    // ancestors, which follow an exponential distribution
    public void genIntervals() {
        int in;
        double t;
	int epoch;

        t = 0.0;

        for (in = sampleSize; in > 1; in--) {
            t += expVar(2 * history.getTheta(0) / (in * (in - 1.0)));
            tree[2 * sampleSize - in].time = t;
        }

	// This complex function stretches or contracts simulated
	// branch lengths, which were generated under the assumption
	// of constant population size, to take population size changes
	// into account. Population increases stretch the simulated
	// branches, population descreases contract them.
        for (in = sampleSize; in < 2 * sampleSize - 1; in++) {
	    epoch = 0;
	    while (tree[in].time > history.getTau(epoch)) {
                tree[in].time = (history.getTheta(epoch+1) /history.getTheta(epoch)) * (tree[in].time - history.getTau(epoch)) + history.getTau(epoch);
		epoch++;
	    }
        }
    }

    // Generate the number of mutations between a node and its
    // ancestor.
    public void genMutations() {
	int i;
	int node;
	double time = 0.0;

       	for (node = 0; node < 2 * sampleSize - 2; node++) {

	    if (tree[node].ancestor != null) {

		time = ((tree[node].ancestor).time)
		    - (tree[node]).time;

		tree[node].nMuts = poiVar(time / 2.0);

	    }
	}
    }

    // Generate a poisson variate
    public int poiVar(double expected) {
        double limit = Math.exp(-1 * expected);
        double product = Math.random();
        int count;

        for ( count = 0; product > limit; count++ )
            product *= Math.random();

        return(count);
    }

    // Generate an exponential variate
    public double expVar(double expected) {
        return (-expected * Math.log(1.0 - Math.random()));
    }

    // Print a human readable representation of the tree
    public void printTopology() {
	int i;

	for (i = 0; i< 2 * sampleSize - 1; i++) {
	    System.out.print(i + ") ");
	    System.out.println(tree[i]);
	}

    }
}
